"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var fs = require("fs-extra");
var graphql_tools_1 = require("graphql-tools");
var STRING_TYPES = [
    'CHAR',
    'VARCHAR',
    'BINARY',
    'VARBINARY',
    'BLOB',
    'TEXT',
    'ENUM',
    'SET',
];
var sqlAST = {};
function addTable(table) {
    sqlAST[table.name] = table;
}
exports.addTable = addTable;
function addColumn(tableName, column) {
    var _a;
    sqlAST[tableName] = __assign(__assign({}, sqlAST[tableName]), { name: tableName, columns: __assign(__assign({}, (sqlAST[tableName] || {}).columns), (_a = {}, _a[column.name] = column, _a)) });
}
exports.addColumn = addColumn;
function makeMysqlSchema(options) {
    sqlAST = {};
    var outputFilepath = options.outputFilepath;
    var databaseName = options.databaseName;
    var tablePrefix = options.tablePrefix || '';
    delete options.outputFilepath;
    delete options.databaseName;
    graphql_tools_1.makeExecutableSchema(options);
    setDefaults();
    gatherIndices();
    renderCreateSchemaScript(databaseName, tablePrefix, outputFilepath);
}
exports.makeMysqlSchema = makeMysqlSchema;
function setDefaults() {
    forEachTableDo(function (table) {
        forEachColumnDo(table, function (column) {
            if (column.primary && column.nullable) {
                emitError(table.name, column.name, '@primary is not allowed with @nullable.');
            }
            if (!column.type) {
                if (column.graphQLType === 'String') {
                    emitError(table.name, column.name, "String types must be explicitly defined using @type.");
                }
                else if (column.graphQLType === 'Int') {
                    column.type = 'INT';
                }
                else if (column.graphQLType === 'Float') {
                    column.type = 'DECIMAL';
                }
                else if (column.graphQLType === 'Boolean') {
                    column.type = 'BOOLEAN';
                }
                else if (column.graphQLType === 'JSON') {
                    column.type = 'JSON';
                }
                else {
                    emitError(table.name, column.name, "A default SQL type cannot be generated for GraphQL type " + column.graphQLType);
                }
            }
            column.type = (column.type || '').toUpperCase();
            if (column.auto) {
                // @ts-ignore
                if (!column.type.includes('INT') && column.type !== 'SERIAL') {
                    emitError(table.name, column.name, "Column with \"auto\" must have INT or SERIAL type.");
                }
                if (column.default) {
                    emitError(table.name, column.name, '"default" is not allowed with "auto".');
                }
                if (column.unicode) {
                    emitError(table.name, column.name, '"unicode" is not allowed with "auto".');
                }
                if (column.generated) {
                    emitError(table.name, column.name, '"generated" is not allowed with "auto".');
                }
            }
            if (column.unicode) {
                // @ts-ignore
                if (!STRING_TYPES.includes(stripLength(column.type))) {
                    emitError(table.name, column.name, '@unicode can only be applied to a string type.');
                }
                if (!table.unicode) {
                    table.unicode = true;
                }
            }
            if (column.generated && column.default) {
                emitError(table.name, column.name, '"default" is not allowed with "generated".');
            }
            if (column.generated && column.generated.startsWith('ALWAYS AS')) {
                column.generated = "GENERATED " + column.generated;
            }
            if (column.generated &&
                !column.generated.startsWith('GENERATED') &&
                !column.generated.startsWith('AS')) {
                column.generated = "AS " + column.generated;
            }
        });
    });
}
function gatherIndices() {
    forEachTableDo(function (table) {
        forEachColumnDo(table, function (column) {
            if (column.primary) {
                if (!!table.primaryIndex) {
                    emitError(table.name, column.name, 'More than one column is marked as the primary index.');
                }
                else {
                    table.primaryIndex = column;
                }
            }
            else if (column.index) {
                table.secondaryIndices = table.secondaryIndices || [];
                table.secondaryIndices.push(column);
            }
        });
        if (!table.primaryIndex) {
            emitError(table.name, '', "Table " + table.name + " does not have a primary index.");
        }
    });
}
function renderCreateSchemaScript(databaseName, tablePrefix, outputFilepath) {
    // console.log('Creating script from:', JSON.stringify(sqlAST, null, '  '))
    var tableDefinitions = [];
    forEachTableDo(function (table) {
        var columnDefinitions = [];
        forEachColumnDo(table, function (column) {
            var unicodeClause = !!column.unicode
                ? 'CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci '
                : '';
            var generatedClause = !!column.generated ? column.generated + " " : '';
            var nullClause = !!column.nullable ? 'NULL ' : 'NOT NULL ';
            var defaultClause = !!column.default ? "DEFAULT " + column.default + " " : '';
            var autoClause = !!column.auto && column.type !== 'SERIAL' ? 'AUTO_INCREMENT ' : '';
            var uniqueClause = !!column.unique ? "UNIQUE " : '';
            columnDefinitions.push(("`" + column.name + "` " + column.type + " " + unicodeClause + generatedClause + nullClause + defaultClause + autoClause + uniqueClause).trim());
        });
        var primaryKeyName = (table.primaryIndex && table.primaryIndex.name) || '';
        var indexDefinitions = (table.secondaryIndices || []).map(function (column) {
            return "INDEX `" + column.name.toUpperCase() + "INDEX` (`" + column.name + "` ASC)";
        });
        if (indexDefinitions.length > 0) {
            indexDefinitions = [''].concat(indexDefinitions);
        }
        var constraints = table.constraints ? ',\n  ' + table.constraints : '';
        // let constraints = (table.constraints || '').split(/,\s*/).join(',\n  ')
        // constraints = constraints ? ',\n  ' + constraints : ''
        var unicodeModifier = table.unicode
            ? ' CHARACTER SET utf8mb4 COLLATE utf8mb4_unicode_ci'
            : '';
        tableDefinitions.push("CREATE TABLE `" + databaseName + "`.`" + tablePrefix + table.name + "` (\n  " + columnDefinitions.join(',\n  ') + ",\n  PRIMARY KEY (`" + primaryKeyName + "`)" + indexDefinitions.join(',\n  ') + constraints + "\n)" + unicodeModifier + ";");
    });
    fs.outputFileSync("" + outputFilepath, tableDefinitions.join('\n\n'));
}
function forEachTableDo(foo) {
    for (var tableName in sqlAST) {
        if (!sqlAST.hasOwnProperty(tableName))
            continue;
        var table = sqlAST[tableName];
        foo(table);
    }
}
function forEachColumnDo(table, foo) {
    for (var columnName in table.columns) {
        if (!table.columns.hasOwnProperty(columnName))
            continue;
        var column = table.columns[columnName];
        foo(column);
    }
}
function emitError(tableName, columnName, message) {
    throw new Error("[" + tableName + " : " + columnName + "]: " + message);
}
function stripLength(type) {
    return type.split('(')[0];
}
